/***
 * Cecilia ADL Compiler
 * Copyright (C) 2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.composite.c.components;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.cecilia.adl.AbstractInstantiationTask;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ClientInterfaceForEach;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Visitor component that builds the code for the call to the
 * constructor/destructor of static component instance.
 */
public class ComponentConstructorVisitor
    extends
      org.objectweb.fractal.cecilia.primitive.thinkMC.components.ComponentConstructorVisitor {

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  @Override
  protected Component createConstructorDeclarationTask(
      final ComponentContainer container) throws TaskException {
    final org.objectweb.fractal.adl.components.Component[] subComps = container
        .getComponents();
    return taskFactoryItf.newPrimitiveTask(
        new ComponentConstructorDeclarationTask(subComps), container, Arrays
            .asList(subComps));
  }

  @Override
  protected Component createConstructorCallTask(
      final ComponentContainer container) throws TaskException {
    final org.objectweb.fractal.adl.components.Component[] subComps = container
        .getComponents();
    return taskFactoryItf.newPrimitiveTask(new ComponentConstructorCallTask(
        subComps), container, Arrays.asList(subComps));
  }

  @Override
  protected Component createDestructorCallTask(
      final ComponentContainer container) throws TaskException {
    final org.objectweb.fractal.adl.components.Component[] subComps = container
        .getComponents();
    return taskFactoryItf.newPrimitiveTask(new ComponentDestructorCallTask(
        subComps), container, Arrays.asList(subComps));
  }

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  @TaskParameters({"componentNode", "subComponents"})
  @ServerInterfaces(@ServerInterface(name = "component-constructor-declaration-provider", signature = SourceCodeProvider.class, record = "role:constructorDeclaration, id:%", parameters = "componentNode"))
  public static class ComponentConstructorDeclarationTask
      extends
        AbstractInstantiationTask {

    protected final org.objectweb.fractal.adl.components.Component[] subComponents;

    // -------------------------------------------------------------------------
    // Task client interfaces
    // -------------------------------------------------------------------------

    /**
     * The {@link SourceCodeProvider} task for each sub component that contains
     * the declarations necessary to call the constructor of the component.
     */
    @ClientInterfaceForEach(prefix = "subComp-constructor-declaration", contingency = TypeFactory.COLLECTION, signature = SourceCodeProvider.class, iterable = "subComponents", record = "role:subComponentConstructorDeclaration, id:%", parameters = "subComponents.element")
    public final Map<Component, SourceCodeProvider>                  subComponentConstructorDeclarationItfs = new HashMap<Component, SourceCodeProvider>();

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param subComponents the sub component nodes.
     */
    public ComponentConstructorDeclarationTask(
        final org.objectweb.fractal.adl.components.Component[] subComponents) {
      this.subComponents = subComponents;
    }

    // -------------------------------------------------------------------------
    // Implementation of abstract methods of AbstractInstantiationTask
    // -------------------------------------------------------------------------
    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter();
      for (final org.objectweb.fractal.adl.components.Component subComponent : subComponents) {
        final SourceCodeProvider provider = subComponentConstructorDeclarationItfs
            .get(subComponent);
        if (provider != null) cw.append(provider.getSourceCode());
      }
      return cw.toString();
    }
  }

  @TaskParameters({"componentNode", "subComponents"})
  @ServerInterfaces(@ServerInterface(name = "component-constructor-call-provider", signature = SourceCodeProvider.class, record = "role:constructorCall, id:%", parameters = "componentNode"))
  public static class ComponentConstructorCallTask
      extends
        AbstractInstantiationTask {

    protected final org.objectweb.fractal.adl.components.Component[] subComponents;

    // -------------------------------------------------------------------------
    // Task client interfaces
    // -------------------------------------------------------------------------

    /**
     * The {@link SourceCodeProvider} task for each sub component that contains
     * the code that call the constructor of the component.
     */
    @ClientInterfaceForEach(prefix = "subComp-constructor-call", contingency = TypeFactory.COLLECTION, signature = SourceCodeProvider.class, iterable = "subComponents", record = "role:subComponentConstructorCall, id:%", parameters = "subComponents.element")
    public final Map<Component, SourceCodeProvider>                  subComponentConstructorCallItfs = new HashMap<Component, SourceCodeProvider>();

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param subComponents the sub component nodes.
     */
    public ComponentConstructorCallTask(
        final org.objectweb.fractal.adl.components.Component[] subComponents) {
      this.subComponents = subComponents;
    }

    // -------------------------------------------------------------------------
    // Implementation of abstract methods of AbstractInstantiationTask
    // -------------------------------------------------------------------------
    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter();
      for (final org.objectweb.fractal.adl.components.Component subComponent : subComponents) {
        final SourceCodeProvider provider = subComponentConstructorCallItfs
            .get(subComponent);
        if (provider != null) cw.append(provider.getSourceCode());
      }
      return cw.toString();
    }
  }

  @TaskParameters({"componentNode", "subComponents"})
  @ServerInterfaces(@ServerInterface(name = "component-destructor-call-provider", signature = SourceCodeProvider.class, record = "role:destructorCall, id:%", parameters = "componentNode"))
  public static class ComponentDestructorCallTask
      extends
        AbstractInstantiationTask {

    protected final org.objectweb.fractal.adl.components.Component[] subComponents;

    // -------------------------------------------------------------------------
    // Task client interfaces
    // -------------------------------------------------------------------------

    /**
     * The {@link SourceCodeProvider} task for each sub component that contains
     * the code that call the destructor of the component.
     */
    @ClientInterfaceForEach(prefix = "subComp-destructor-call", contingency = TypeFactory.COLLECTION, signature = SourceCodeProvider.class, iterable = "subComponents", record = "role:subComponentDestructorCall, id:%", parameters = "subComponents.element")
    public final Map<Component, SourceCodeProvider>                  subComponentDestructorCallItfs = new HashMap<Component, SourceCodeProvider>();

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param subComponents the sub component nodes.
     */
    public ComponentDestructorCallTask(
        final org.objectweb.fractal.adl.components.Component[] subComponents) {
      this.subComponents = subComponents;
    }

    // -------------------------------------------------------------------------
    // Implementation of abstract methods of AbstractInstantiationTask
    // -------------------------------------------------------------------------
    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter();
      for (final org.objectweb.fractal.adl.components.Component subComponent : subComponents) {
        final SourceCodeProvider provider = subComponentDestructorCallItfs
            .get(subComponent);
        if (provider != null) cw.append(provider.getSourceCode());
      }
      return cw.toString();
    }
  }
}
